use std::{
    collections::HashMap,
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
};

#[derive(PartialEq, Eq, Debug)]
enum RequestMethods {
    NULL = -1, // This is only to initialise the struct
    GET,
    HEAD,
}

#[derive(Debug)]
struct RequestLine {
    method: RequestMethods,
    target: String,
    version: String,
}

impl RequestLine {
    pub fn new() -> Self {
        Self {
            method: RequestMethods::NULL,
            target: String::new(),
            version: String::new(),
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc9110#name-methods
    pub fn is_valid_method(method: &str) -> bool {
        if method.trim().is_empty() {
            return false;
        }

        // Only GET and HEAD are required, the rest is optional
        if ["GET", "HEAD"].contains(&method.trim()) {
            return true;
        } else {
            return false;
        }
    }

    // TODO: make the checks less shit and actually correct
    // https://datatracker.ietf.org/doc/html/rfc9112#name-request-target
    pub fn is_valid_target(target: &str) -> bool {
        if target.trim().is_empty() {
            return false;
        }

        // origin-form
        if target.starts_with("/") {
            if target.contains("?") && target.split("?").count() != 2 {
                return false;
            }
            return true;
        }

        // absolute-form
        if target.starts_with("http://") {
            return true;
        }

        // authority-form
        if target.contains(":") && target.split(":").count() == 2 {
            return true;
        }

        // asterisk-form
        if target.trim() == "*" {
            return true;
        }

        return false;
    }

    pub fn is_valid_version(version: &str) -> bool {
        if version.trim().is_empty() {
            return false;
        }

        let http_name = version.trim();

        if http_name.starts_with("HTTP/") {
            let version_numbers = http_name.trim_start_matches("HTTP/");
            let version_numbers = version_numbers.split(".").collect::<Vec<&str>>();
            if version_numbers.len() != 2 {
                return false;
            }
            let major = match version_numbers[0].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };
            let minor = match version_numbers[1].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };

            if major <= 9 && minor <= 9 {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
}

fn parse_start_line(input: &str) -> Result<RequestLine, Vec<u8>> {
    let mut response_field_lines: HashMap<String, String> = HashMap::new();
    let mut response_body: Vec<u8> = vec![];

    if input.ends_with(" ") {
        for byte in b"There is whitespace between the start-line and the first field-line\r\n" {
            response_body.push(*byte);
        }
        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 400 Bad Request",
            response_field_lines,
            response_body,
        ));
    }

    let mut start_line = RequestLine::new();
    let vec = input.trim().split_ascii_whitespace().collect::<Vec<&str>>();

    let body = format!(
        "The start-line has an incorrect amount of items. Got the value: {}\r\n",
        vec.len()
    );

    if vec.len() != 3 {
        for byte in body.as_bytes() {
            response_body.push(*byte);
        }
        response_field_lines.insert(
            String::from("Content-Length"),
            response_body.len().to_string(),
        );
        response_field_lines.insert(String::from("Content-Type"), String::from("text/plain"));

        return Err(response_builder(
            RequestMethods::GET,
            "HTTP/1.1 400 Bad Request",
            response_field_lines,
            response_body,
        ));
    }

    // start_line.method will remain RequestMethods::NULL if it is not supported.
    let method = vec[0];
    if RequestLine::is_valid_method(&method) {
        // TODO: Change to a switch-case if I ever support more methods

        if method == String::from("GET") {
            start_line.method = RequestMethods::GET;
        }
        if method == String::from("HEAD") {
            start_line.method = RequestMethods::HEAD;
        }
    }

    let target = vec[1];
    if RequestLine::is_valid_target(&target) {
        start_line.target = target.to_string();
    }

    let version = vec[2];
    if RequestLine::is_valid_version(&version) {
        if version == "HTTP/1.1" || version == "HTTP/1.0" {
            start_line.version = version.to_string();
        }
    }

    return Ok(start_line);
}

fn parse_field_lines(reader: &mut BufReader<&mut TcpStream>) -> Option<HashMap<String, String>> {
    let mut line: String;
    let mut field_lines: HashMap<String, String> = HashMap::new();
    let mut is_first_line = true;

    // Read field-lines till I hit an empty line
    loop {
        line = String::new();
        reader.read_line(&mut line).unwrap();

        if line.starts_with(" ") && is_first_line {
            return None;
        }

        if !line.ends_with("\r\n") {
            return None;
        }

        if line.trim().is_empty() {
            break;
        }

        let field_line = line.split_once(":").expect("Shits fucked: {}");

        // Check if client has send more than one Host line
        if field_lines.contains_key(&String::from("Host")) && field_line.0 == "Host" {
            return None;
        }

        field_lines.insert(field_line.0.to_owned(), field_line.1.trim().to_owned());

        is_first_line = false;
    }

    if !field_lines.contains_key(&String::from("Host")) {
        return None;
    }

    return Some(field_lines);
}

fn response_builder(
    method: RequestMethods,
    status_line: &str,
    field_lines: HashMap<String, String>,
    body: Vec<u8>,
) -> Vec<u8> {
    let mut response: Vec<u8> = vec![];

    // TODO: replaces with eiter Option<T> or Result<T, T>
    if status_line.is_empty() {
        return response;
    }

    for byte in status_line.as_bytes().iter() {
        response.push(*byte);
    }
    response.push(b'\r');
    response.push(b'\n');

    if !field_lines.is_empty() {
        for field_line in field_lines.iter() {
            for byte in field_line.0.as_bytes().iter() {
                response.push(*byte);
            }

            response.push(b':');
            response.push(b' ');

            for byte in field_line.1.as_bytes().iter() {
                response.push(*byte);
            }

            response.push(b'\r');
            response.push(b'\n');
        }
    }

    // Mandatory empty line between header and body
    response.push(b'\r');
    response.push(b'\n');

    if body.is_empty() {
        return response;
    }

    if method != RequestMethods::HEAD && method != RequestMethods::NULL {
        for byte in body.iter() {
            response.push(*byte);
        }
        response.push(b'\r');
        response.push(b'\n');
    }

    return response;
}

fn handle_request(mut stream: TcpStream) {
    let mut line = String::new();
    let mut reader = BufReader::new(&mut stream);

    // Request can have one or many empty lines preceding the start-line and I will ignore these
    loop {
        match reader.read_line(&mut line) {
            Ok(val) => {
                if val > 2 {
                    break;
                }
            }
            // TODO: Replace with stream.write_all
            Err(err) => {
                eprintln!("{err}");
                return;
            }
        };

        if line != "\r\n" {
            continue;
        }
    }

    let start_line = match parse_start_line(&line) {
        Ok(val) => val,
        Err(response) => {
            stream.write_all(&response).unwrap();
            return;
        }
    };
    dbg!(&start_line);

    if start_line.method == RequestMethods::NULL {
        stream
            .write_all(b"HTTP/1.1 501 Not Implemented\r\n\r\n")
            .unwrap();
        return;
    }

    let field_lines = match parse_field_lines(&mut reader) {
        Some(val) => val,
        None => {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nInvalid Header")
                .unwrap();
            return;
        }
    };
    dbg!(&field_lines);

    // TODO: Read the body
    // let mut body: Vec<u8> = vec![];
    // reader.read_to_end(&mut body).unwrap();
    // dbg!(&body);
    // TODO: Act upon the request

    stream
        .write_all(b"HTTP/1.1 200 OK\r\n\r\nThis is the server speaking\r\n")
        .unwrap();
}

fn main() -> Result<(), std::io::Error> {
    let listener = TcpListener::bind("0.0.0.0:8080")?;

    println!("Server started");

    for stream in listener.incoming() {
        let stream = stream?;
        handle_request(stream)
    }
    Ok(())
}
